import sys
from pathlib import Path
import matplotlib.pyplot as plt
from typing import List, Tuple
import argparse


def parse_args(args: List[str]) -> Tuple[argparse.Namespace, dict]:
    parser = argparse.ArgumentParser("Embed binary single cell/nucleus sequencing data using k-mer fragmentation over "
                                     "single and across species.")
    parser.add_argument(
        "--setup_file", type=str, required=True,
        help="Path to setup file in framework-specific .embed format."
    )
    parser.add_argument(
        "--k", type=int, default=5,
        help="Length of sequences."
    )
    parser.add_argument(
        "--norm", type=str, default="centered_sum",
        help="Normalization for KMer matrices. Choose between none | max_norm | sum_norm | centered_sum | "
             "centered_max | centered_uniform_sum | centered_uniform_max."
    )
    parser.add_argument(
        "--max_region_size", type=int, default=701,
        help="Maximum peak size that is included for KMer creation."
    )
    parser.add_argument(
        "--n_policy", type=str, default="remove",
        help="How to deal with indeterministic sequence parts. Choose between remove | replace | keep."
    )
    parser.add_argument(
        "--mask_rep", action="store_true", dest="mask_rep",
        help="If set, mask repetitive sequences and don't consider them for KMer creation."
    )
    parser.add_argument(
        "--equalize_counter", action="store_true", dest="equalize_counter",
        help="If set, treat forward and reverse complement kmer as the same kmer, e.g. `ACGC` counts are "
            "added to `GCGT` counts."
    )
    parser.add_argument(
        "--verbosity", type=int, default=2,
        help="Verbosity level. Set to -1 for no output, 0 for only printing warnings and 1 for complete output."
    )
    parser.add_argument(
        "--data_path", type=str, default="./",
        help="Data output directory."
    )
    parser.add_argument(
        "--fig_path", type=str, default="./",
        help="Figure output directory."
    )
    parser.add_argument(
        "--save_prefix", type=str, default="",
        help="Save prefix added to saved file names."
    )
    parser.add_argument(
        "--correct_gc", action="store_true", dest="correct_gc",
        help="If set, correct for GC distribution. The GC content per cell must then match the global GC "
            "content in all peaks."
    )
    parser.add_argument(
        "--n_jobs", type=int, default=8,
        help="Number of CPUs used."
    )
    args, addition_arg_list = parser.parse_known_args(args)
    addition_args = {opt_arg.split("=")[0].replace("--", ""): opt_arg.split("=")[1]
                     for opt_arg in addition_arg_list}
    return args, addition_args


def main():
    args_list = sys.argv[1:]
    args, optional_args = parse_args(args_list)

    from spyce.utils import load_specifications
    from spyce.kmerMatrix import KMerCollection
    setup_file_path = args.setup_file
    k = args.k
    normalization = args.norm
    n_policy = args.n_policy
    mask_rep = args.mask_rep
    equalize_counter = args.equalize_counter
    verbosity = args.verbosity
    data_path = args.data_path
    fig_path = args.fig_path
    save_prefix = args.save_prefix
    correct_gc = args.correct_gc
    max_region_size = args.max_region_size
    n_jobs = args.n_jobs

    Path(data_path).mkdir(exist_ok=True, parents=True)
    Path(fig_path).mkdir(exist_ok=True, parents=True)
    setup_dict = load_specifications(setup_file_path)
    kmer_collect = KMerCollection(
        setup_dict=setup_dict,
        k=k,
        verbosity=verbosity,
        n_jobs=n_jobs,
        normalization=normalization,
        n_policy=n_policy,
        correct_gc=correct_gc,
        equalize_counter=equalize_counter,
        mask_rep=mask_rep,
        max_region_size=max_region_size
    )
    if verbosity > 0:
        print("Finished creating KMer collection, save to file.")
    save_path = "%s/%skmer_collection_obj.pkl" % (data_path, save_prefix)
    kmer_collect.save(path=save_path, save_prefix=save_prefix)


if __name__ == '__main__':
    main()




import sys
from pathlib import Path
import matplotlib.pyplot as plt
from typing import List, Tuple
import argparse

from spyce.utils import load_specifications
from spyce.kmerMatrix import KMerCollection
from spyce.plotting import plot_umap


def parse_args(args: List[str]) -> Tuple[argparse.Namespace, dict]:
    parser = argparse.ArgumentParser("Embed binary single cell/nucleus sequencing data using k-mer fragmentation over "
                                     "single and across species.")
    parser.add_argument(
        "--setup_file", type=str, required=True,
        help="Path to setup file in framework-specific .embed format."
    )
    parser.add_argument(
        "--k", type=int, default=5,
        help="Length of sequences."
    )
    parser.add_argument(
        "--verbosity", type=int, default=2,
        help="Verbosity level. Set to -1 for no output, 0 for only printing warnings and 1 for complete output."
    )
    parser.add_argument(
        "--dr", type=str,
        help="Dimensionality reduction method that is used. It is recommended to use a dimensionality reduction "
             "when k > 5. If none passed, none is performed."
    )
    parser.add_argument(
        "--clustering_per_sample", type=str,
        help="If set, use this algorithm to cluster each sample individually according to its k-mer histogram. "
             "Choose between (not case-sensitive) leiden | KMeans | DBSCAN | OPTICS | hierarchical_KMeans. "
             "Their behavior can be specified by passing additional cml-parameters (see documentation)."
    )
    parser.add_argument(
        "--sample_cluster_name", type=str, default="clustering",
        help="Name used for saving the clustering result in object."
    )
    parser.add_argument(
        "--clustering_total", type=str,
        help="If set, use this algorithm to cluster the entire data set (e.g. across conditions or species) using the "
             "combined k-mer histogram. "
             "Choose between (not case-sensitive) leiden | KMeans | DBSCAN | OPTICS | hierarchical_KMeans. "
             "Their behavior can be specified by passing additional cml-parameters (see documentation)."
    )
    parser.add_argument(
        "--total_cluster_name", type=str, default="clustering",
        help="Name used for saving the clustering result in object."
    )
    parser.add_argument(
        "--seed", type=int,
        help="If set, use a seed to create reproducible results. Note that not all clustering algorithms accept seeds."
    )
    parser.add_argument(
        "--n_umap_neighbors", type=int, default=15,
        help="Number of nearest neighbors used for UMAP embedding."
    )
    parser.add_argument(
        "--umap_metric", type=str, default="euclidean",
        help="Metric used for calculating UMAP. See https://umap-learn.readthedocs.io/en/latest/parameters.html#metric."
    )
    parser.add_argument(
        "--umap_min_dist", type=float, default=.1,
        help="Minimum distance used for UMAP embedding."
    )
    parser.add_argument(
        "--n_umap_comp", type=int, default=2,
        help="Dimension of UMAP manifold. "
             "For plotting this value needs to be 1 <= n_umap_comp <= 3 or will otherwise fail."
    )
    parser.add_argument(
        "--color_key", action="append", type=str,
        help="Set according to which values in UMAP embedding should be colored during plotting. This value can be "
             "species | matrix | the clustering used (pass total_cluster_name or sample_cluster_name) | "
             "additional cell values that were either present in h5ad files defined in setup_file or in "
             "separate tables defined in setup_file (see documentation). Default is matrix."
    )
    parser.add_argument(
        "--cmap", type=str, default="nipy_spectral",
        help="Type of color map used for plotting. "
             "See https://matplotlib.org/stable/users/explain/colors/colormaps.html"
    )
    parser.add_argument(
        "--data_path", type=str, default="./",
        help="Data output directory."
    )
    parser.add_argument(
        "--fig_path", type=str, default="./",
        help="Figure output directory."
    )
    parser.add_argument(
        "--save_prefix", type=str, default="",
        help="Save prefix added to saved file names."
    )
    parser.add_argument(
        "--n_jobs", type=int, default=8,
        help="Number of CPUs used."
    )
    args, addition_arg_list = parser.parse_known_args(args)
    addition_args = {opt_arg.split("=")[0].replace("--", ""): opt_arg.split("=")[1]
                     for opt_arg in addition_arg_list}
    return args, addition_args


def main(args_list: List[str]):
    args, optional_args = parse_args(args_list)
    setup_file_path = args.setup_file
    k = args.k
    verbosity = args.verbosity
    clustering_per_sample = args.clustering_per_sample
    clustering_total = args.clustering_total
    seed = args.seed
    clustering_name_per_sample = args.sample_cluster_name
    clustering_name_total = args.total_cluster_name
    n_umap_neighbors = args.n_umap_neighbors
    dr = args.dr
    umap_metric = args.umap_metric
    umap_min_dist = args.umap_min_dist
    n_umap_comp = args.n_umap_comp
    color_key = args.color_key if args.color_key is not None else ["matrix"]
    cmap = args.cmap
    data_path = args.data_path
    fig_path = args.fig_path
    save_prefix = args.save_prefix
    n_jobs = args.n_jobs

    Path(data_path).mkdir(exist_ok=True, parents=True)
    Path(fig_path).mkdir(exist_ok=True, parents=True)
    setup_dict = load_specifications(setup_file_path)
    kmer_collect = KMerCollection(setup_dict=setup_dict, k=k, verbosity=verbosity, n_jobs=n_jobs)

    if dr is not None:
        if verbosity > 0:
            print("Perform dimensionality reduction.")
        kmer_collect.reduce_dimensionality(
            algorithm=dr,
            seed=seed,
            save_name=dr,
            verbosity_indent="\t",
            n_jobs=n_jobs
        )

    if clustering_per_sample is not None:
        if verbosity > 0:
            print("Perform clustering per data sample")
        for kmer_mat in kmer_collect.kmer_mat_list:
            kmer_mat.cluster(
                algorithm=clustering_per_sample,
                seed=seed,
                save_name=clustering_name_per_sample,
                verbosity_indent="\t",
                n_jobs=n_jobs,
                **optional_args
            )

    if clustering_total is not None:
        if verbosity > 0:
            print("Perform clustering over the entire dataset")
        kmer_collect.run_clustering(
            algorithm=clustering_total,
            dr_name=dr,
            seed=seed,
            save_name=clustering_name_total,
            verbosity_indent="\t",
            n_jobs=n_jobs,
            **optional_args
        )

    if verbosity > 0:
        print("Compute UMAP of entire dataset.")

    kmer_collect.umap(
        n_neighbors=n_umap_neighbors,
        dr_name=dr,
        metric=umap_metric,
        min_dist=umap_min_dist,
        n_components=n_umap_comp,
        verbosity_indent="\t"
    )

    if data_path is not None:
        save_path = "%s/%skmer_collection_obj.pkl" % (data_path, save_prefix)
        if verbosity > 0:
            print("Save KMerCollection object under %s" % save_path)
        kmer_collect.save(path=save_path)

    if verbosity > 0:
        print("Start plotting KMer embedding over entire data set")

    for ck in color_key:
        if verbosity > 0:
            print("\tColor plot w.r.t. %s" % ck)
        fig, ax = plot_umap(
            kmer_obj=kmer_collect,
            ck=ck,
            randomize=True,
            title=ck,
            cmap=cmap,
            verbosity=verbosity
        )
        fig.savefig("%s/%s%s.pdf" % (fig_path, save_prefix, ck), dpi=300)
        plt.close(fig)


if __name__ == "__main__":
    main(sys.argv[1:])





#: Letters in the DNA code.
DNA_ALPHABET = ["A", "C", "G", "T"]
#: Letters in the RNA code.
RNA_ALPHABET = ["A", "C", "G", "U"]
#: DNA colors for plotting. There is no known convention for colour code, see discussion here https://www.biostars.org/p/171056/
DNA_ALPHABET_COLOR = {"A": "tab:blue", "C": "tab:orange", "G": "tab:green", "T": "tab:red"}
#: Maximum value for k until which sPYce will use a dense matrix representation. Note that most k-mers are normally present, and using a sparse matrix does not reduce memory when using k too low.
MAX_K_DENSE = 15
#: Strings that are used for GC content correction.
GC_CONTENT = ["C", "G"]
GC_TUPLE = ["CC", "CG", "GC", "GG"]
#: URL where to report a bug.
BUG_REPORT = "https://gitlab.pasteur.fr/cofugeno/spyce/-/issues"



from __future__ import annotations

from typing import Dict, List, Tuple
from tqdm import tqdm
from multiprocessing.pool import AsyncResult
import time
import warnings
from sklearn.neighbors import KDTree
from sklearn.metrics.pairwise import euclidean_distances
from scipy import sparse as sp
import numpy as np
from itertools import product
from Bio.Seq import Seq
from pandas import Series as pd_Series
from polars import Series as pl_Series

from spyce.constants import DNA_ALPHABET
from spyce.dataType import DataLoadingStructure, KMerHistogramObject


class TopicNode(DataLoadingStructure):
    """
    Node class for parsing framework-specific .embed files, which provide hierarchical information by indent
    Instances of this class are nodes in a topic tree that represents the hierarchical structure of the .embed files
    
    :param str indented_line: entire line of file as string including indents
    """
    def __init__(self, indented_line: str):
        self.children = []
        self.level = len(indented_line) - len(indented_line.lstrip())
        self.text = indented_line.strip()

    def add_children(self, topic_nodes: List):
        """
        Add topic children to node
        
        :param List topic_nodes: Topic nodes that are added as children
        :return: None
        """
        childlevel = topic_nodes[0].level
        while topic_nodes:
            tp = topic_nodes.pop(0)
            if tp.level == childlevel:  # add node as a child
                self.children.append(tp)
            elif tp.level > childlevel:  # add nodes as grandchildren of the last child
                topic_nodes.insert(0, tp)
                self.children[-1].add_children(topic_nodes)
            elif tp.level <= self.level:  # this node is a sibling, no more children
                topic_nodes.insert(0, tp)
                return

    def as_dict(self) -> Dict | str:
        """
        Convert hierarchical tree structure of which this node is root as dict
        
        :return Dict | str: tree structure as dict. If leaf node return string
        """
        if len(self.children) > 1:
            child_dict = {} 
            child_list = []
            for tp in self.children:
                tp_dict = tp.as_dict()
                try:
                    child_dict = dict(**child_dict, **tp_dict)
                except TypeError:
                    child_list.append(tp_dict)
            return {self.text: child_dict} if len(child_list) == 0 else {self.text: child_list}
        elif len(self.children) == 1:
            return {self.text: self.children[0].as_dict()}
        else:
            return self.text


def load_specifications(path: str) -> Dict:
    """
    Load an `.embed` file, parse to hierarchical tree, and convert to dict
    
    :param str path: path to .embed file
    :return Dict: python dict with hierarchical information
    """
    root_topic = TopicNode("root")
    with open(path, "r") as spec_file:
        root_topic.add_children([TopicNode(line) for line in spec_file.read().splitlines() if line.strip()])
    return root_topic.as_dict()["root"]


def seq_to_int(seq: str, alphabet: List[str]) -> int:
    """
    Convert a sequence of letters to an integer value given a full alphabet
    
    :param str seq: sequence string, e.g. ACCAGTA
    :param List[str] alphabet: List with single characters representing alphabet
    :return int: integer representing sequence
    """
    alphabet = list(sorted(alphabet))
    value = 0
    for i_letter, letter in enumerate(reversed(seq)):
        value += alphabet.index(letter) * len(alphabet)**i_letter

    return value

def nozero_kmer_to_idx(alphabet: List[str] = DNA_ALPHABET, k: int = 4) -> Dict[str, int]:
    """
    To avoid a bias based in the orientation of the reference genome, we treat a k-mer and its
    reverse implement (for example for `k=3`, `AAT` and `ATT` are treated as the same k-mer).
    This creates many values in the kmer histogram that will always be zero. This function
    returns a dictionary with all present k-mers and their indices.

    :param List[str] alphabet: Alphabet used for creating k-mer histogram.
    :param int k: Length of the k-mers
    :return Dict[str, int]: dictionary with (k-mer, index) key:value pairs.
    """
    kmer_idx_dict = {}
    ctr = 0
    for kmer in sorted(product(alphabet, repeat=k)):
        kmer = "".join(kmer)
        rev_kmer =  str(Seq(kmer).reverse_complement())
        save_kmer = kmer if kmer < rev_kmer else rev_kmer
        if save_kmer in kmer_idx_dict:
            continue
        else:
            kmer_idx_dict[save_kmer] = ctr
            ctr += 1
    return kmer_idx_dict


def fetch_async_result(
        job_list: List[Tuple[int | str, AsyncResult]],
        process_bar: tqdm | None = None,
        max_attempt: int | None = 200  # approx 100 secs
    ) -> List:
    """
    Fetch result from asynchronous job when ready.

    :param List[AsyncResult] job_list: list with asynchronous results.
    :param tqdm | None process_bar: Process bar. If none, no process bar is plotted
    :param int | None max_attempt: Maximum number of attempts to fetch result with half a second
        sleep time between each completed iteration through all remaining open jobs.
    :return List: List with results
    """
    processed_jobs = set()
    results = []
    attempts = 0
    while len(job_list) > 0:
        if attempts > max_attempt:
            warnings.warn("Reached limit of %d attempts without results. Continue." % max_attempt)
            break
        # get first in list
        i, async_res = job_list.pop(0)
        # wait when iterated through entire job list and still unfinished jobs
        if i in processed_jobs:
            # attempt to fetch the same
            attempts += 1
            time.sleep(.5)
            processed_jobs = set()
        processed_jobs.add(i)
        # check if ready and fetch
        if async_res.ready():
            attempts = 0
            results.append((i, async_res.get()))
            if process_bar is not None:
                process_bar.update(1)
        else:
            # otherwise add to end of list.
            job_list.append((i, async_res))
    return results


def get_dist_mat(data: np.ndarray) -> np.ndarray:
    return euclidean_distances(data, data)


def get_nn_mat(
        data: np.ndarray | KMerHistogramObject,
        n_neighbors: int | None = 10,
        radius: float | None = None,
        neighbor_batch: int = 128,
        return_distance: bool = True,
        dr_name: str | None = None,
        verbosity: int = 0,
        verbosity_indent: str = ""
) -> sp.csc_matrix:
    """
    Calculate nearest neighbor adjacency matrix.

    :param np.ndarray | KMerHistogramObject data: Input data of size `(#cells x #features)`
    :param int n_neighbors: Number of nearest neighbors per cell.
    :param float | None radius: Use neighborhood radius around each cell rather than k-nearest neighbors.
    :param int neighbor_batch: Number of cells processed at the same time for finding the nearest neighbors.
    :param bool return_distance: Return distance between cell and neighbors instead of adjacency.
    :param str | None dr_name: Name of dimensionality reduction that should be used.
    :param int verbosity: Verbosity levels.
    :param str verbosity_indent: Prefix that is added to the output.
    :return sp.csc_matrix: Sparse adjacency matrix.
    """
    # number of neighbors must be larger than 1, otherwise only returns autoconnections
    if n_neighbors <= 1:
        raise ValueError("Number of neighbors must be at least 2.")
    
    # check whether data is KMerClass and extract data if necessary
    if isinstance(data, KMerHistogramObject):
        if dr_name is not None and dr_name in data.dr:
            data = data.dr[dr_name]
        else:
            data = data.kmer_hist

    # create KD tree
    kdtree = KDTree(data=data)
    adj_mat = sp.lil_matrix((data.shape[0], data.shape[0]), dtype="float")
    cell_kmer_iterator = tqdm(
        range(0, data.shape[0], neighbor_batch),
        desc="%s\tProgress" % verbosity_indent
    ) if verbosity > 0 else range(0, data.shape[0], neighbor_batch)
    # iterate over all cells to create adjacency matrix
    for i_cell_start in cell_kmer_iterator:
        # if based on neighbors, query with number of neighbors
        if radius is None and n_neighbors is not None:
            dist, neighbor_idc = kdtree.query(
                data[i_cell_start:np.minimum(i_cell_start + neighbor_batch, data.shape[0])],
                k=n_neighbors,
                return_distance=True
            )
            cell_idc = np.repeat(
                np.arange(i_cell_start, np.minimum(i_cell_start + neighbor_batch, data.shape[0]), 1),
                n_neighbors 
            )
        # if based on radius, query neighborhood of input data points
        elif radius is not None:
            dist, neighbor_idc = kdtree.query_radius(
                data[i_cell_start:np.minimum(i_cell_start + neighbor_batch, data.shape[0])],
                r=radius,
                return_distance=True
            )
            cell_idc = np.concatenate([
                i * np.ones(len(neighbor_idc[num]))
                for num, i in enumerate(
                    np.arange(i_cell_start, np.minimum(i_cell_start + neighbor_batch, data.shape[0]), 1)
                )
            ])
            dist = np.concatenate(dist)
            neighbor_idc = np.concatenate(neighbor_idc)
        else:
            raise ValueError("Pass either `n_neighbors` or `radius`.")

        # set adjacency matrix. If flag is set, return distance
        adj_mat[cell_idc, neighbor_idc.reshape(-1)] = dist.reshape(-1) if return_distance else 1.
    adj_mat = adj_mat.tocsc()
    return adj_mat


def match_annotations(
        annot_x: List | np.ndarray | pl_Series | pd_Series,
        annot_y: List | np.ndarray | pl_Series | pd_Series
) -> np.ndarray:
    """
    Return mask for matching annotation labels.

    :param List | np.ndarray | pl_Series | pd_Series annot_x: Cell type annotation
    :param List | np.ndarray | pl_Series | pd_Series annot_y: Cell type annotation `annot_x` is compared to
    :return np.ndarray: Mask for matching labels
    """
    def _convert(annot):
        if isinstance(annot, list):
            annot = np.array(annot)
        elif isinstance(annot_x, pl_Series) or isinstance(annot, pd_Series):
            annot = annot_x.to_numpy()
        return annot
    
    annot_x = _convert(annot_x)
    annot_y = _convert(annot_y)
    return annot_x == annot_y

